# Copyright (c) 2019-2020 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

import sympy
from sympy.logic.boolalg import to_cnf, to_dnf
from sympy.logic import POSform, SOPform
from sympy.utilities.lambdify import lambdify

from itertools import chain, product, count, combinations

import networkx as nx

from typing import Any, Tuple, List

from ..lccommon.data_types import Transistor, ChannelType

import logging

logger = logging.getLogger(__name__)


def test_convert_to_dnf():
    """
    Example for finding the DNF of a boolean formula.
    :return:
    """
    a, b, s = sympy.symbols('a b s')

    z = (a & s) | (b & ~s)

    dnf1 = to_dnf(z, simplify=True)
    dnf2 = sympy.simplify_logic(z, form='dnf')

    assert dnf1 == dnf2

    x = z

    def print_tree(expr: sympy.Symbol, depth=0):
        tab = ' ' * depth
        print(tab, expr.func, expr.args)
        for arg in expr.args:
            print_tree(arg, depth + 1)

    # print_tree(x)


def _formula_to_cmos_network(f: sympy.Symbol):
    counter = count(1)

    def _formula_to_pull_network(f: sympy.Symbol, channel_type: ChannelType) -> nx.MultiGraph:
        """ Create the pull-up or pull-down network of a boolean forumla `f` in DNF or CNF.

        :param f: Boolean formula in DNF or CNF.
        :param channel_type:
        :return: Returns a nx.MultiGraph representing the CMOS network. Each edge represents a transistor.
                Each node represents a net. Special nets are 'gnd', 'vdd' and 'z'. 'z' is the output net of the network.
        """

        if isinstance(f, sympy.And) or isinstance(f, sympy.Or):

            graphs = [_formula_to_pull_network(a, channel_type) for a in f.args]

            parallel = isinstance(f, sympy.Or)
            serial = isinstance(f, sympy.And)

            assert serial ^ parallel

            if not parallel:
                # Put sub graphs in series
                a = graphs[0]
                for b in graphs[1:]:
                    joint = next(counter)
                    mapping_a = {'power': joint}
                    mapping_b = {'output': joint}
                    a = nx.relabel_nodes(a, mapping_a)
                    b = nx.relabel_nodes(b, mapping_b)
                    a = nx.compose(a, b)

                graph = a
            elif parallel:
                # Put sub graphs in parallel
                graph = nx.compose_all(graphs)
        else:

            assert isinstance(f, sympy.Not) or isinstance(f, sympy.Symbol), TypeError("Expected And, Or or Symbol.")

            if channel_type == ChannelType.PMOS:
                f = sympy.simplify(sympy.Not(f))

            graph = nx.MultiGraph()
            graph.add_edge('power', 'output', key=(f, channel_type))

        assert graph is not None
        return graph

    normal_form = 'dnf'
    nf = sympy.simplify_logic(f, form=normal_form)
    nf_neg = sympy.simplify_logic(~f, form=normal_form)

    nmos_graph = _formula_to_pull_network(nf_neg, ChannelType.NMOS)
    pmos_graph = _formula_to_pull_network(nf, ChannelType.PMOS)

    nmos_graph = nx.relabel_nodes(nmos_graph, {'power': 'gnd'})
    pmos_graph = nx.relabel_nodes(pmos_graph, {'power': 'vdd'})

    cmos_graph = nx.compose(nmos_graph, pmos_graph)

    return cmos_graph


def _cmos_graph_to_transistors(input_signals, cmos_graph: nx.MultiGraph) -> List[Transistor]:
    """
    Convert a labeled multigraph into the corresponding transistor network.
    :param input_signals:
    :param cmos_graph:
    :return:
    """
    all_nets = set()
    # Create normalized net names
    for source, drain, (gate, channel_type) in cmos_graph.edges(keys=True):
        all_nets.add(source)
        all_nets.add(drain)
        all_nets.add(gate)

    cnt = count(1)
    net_name_mapping = dict()
    for net in all_nets:
        if isinstance(net, str):
            name = net
        elif net in input_signals:
            name = str(net)
        elif net == 'output':
            name = 'output'
        else:
            name = '%d' % (next(cnt))

        net_name_mapping[net] = name

    cmos_graph = nx.relabel_nodes(cmos_graph, net_name_mapping, copy=True)

    # Create transistors
    transistors = []
    for source, drain, (gate, channel_type) in cmos_graph.edges(keys=True):
        t = Transistor(channel_type, source, gate, drain, channel_width=1)
        transistors.append(t)

    return transistors


def synthesize_circuit(f: sympy.Symbol) -> nx.MultiGraph:
    """ Synthesize a CMOS network implementing the boolean formula `f`.
    :param f:
    :return: A nx.MultiGraph representing the transistor network.
    """
    cmos_graph = _formula_to_cmos_network(f)

    # Create cmos networks for all gate inputs which are not just a signal.
    # For example this generates inverted input signals when needed.
    input_signals = set(f.atoms())
    input_sources = set()

    for _, _, (input_formula, channel_type) in cmos_graph.edges(keys=True):
        if input_formula not in input_signals and input_formula not in input_sources:
            input_circuit = _formula_to_cmos_network(input_formula)
            input_sources.add(input_circuit)

    # Merge input networks into full graph.
    cmos_graph = nx.compose_all(list(chain([cmos_graph], input_sources)))

    return cmos_graph


def test_synthesize_circuit():
    ''' Test synthesizing an inverter.
    :return:
    '''
    a = sympy.symbols('a')
    cmos = synthesize_circuit(~a)
    assert len(cmos.edges) == 2


def synthesize_minimal_circuit(z: sympy.Symbol) -> nx.MultiGraph:
    """ Synthesize `z` and `Not(z)` + an inverter and return the circuit with less transistors.
    :param z:
    :return:
    """
    z_inv = ~z

    cmos_graph_inv = synthesize_circuit(z_inv)
    cmos_graph = synthesize_circuit(z)

    if len(cmos_graph.edges) > len(cmos_graph_inv.edges) + 2:
        logger.debug('Construct inverted circuit + inverter: Neg(Neg(%s))', z)
        cmos_graph = cmos_graph_inv
        # Append inverter
        output = sympy.symbols('output_inv')
        inverter = _formula_to_cmos_network(~output)

        cmos_graph = nx.relabel_nodes(cmos_graph, {'output': 'output_inv'})
        cmos_graph = nx.compose(cmos_graph, inverter)

    return cmos_graph


def test_synthesize_minimal_circuit():
    a, b = sympy.symbols('a b')
    z = a & b
    cmos = synthesize_minimal_circuit(z)
    assert len(cmos.edges) == 4 + 2


def synthesize_transistors(f: sympy.Symbol) -> List[Transistor]:
    cmos_graph = synthesize_minimal_circuit(f)
    input_signals = set(f.atoms())
    transistors = _cmos_graph_to_transistors(input_signals, cmos_graph)

    return transistors


def formula_from_string(s: str) -> sympy.Symbol:
    return sympy.parsing.sympy_parser.parse_expr(s)


def test_from_minterms():
    """
    Example of how to get a boolean formula from minterms.
    :return:
    """
    w, x, y, z = sympy.symbols('w x y z')

    minterms = [(0, 0, 0, 1), (0, 0, 1, 1), (0, 1, 1, 1), (1, 0, 1, 1), (1, 1, 1, 1)]
    dontcares = [(0, 0, 0, 0), (0, 0, 1, 0), (0, 1, 0, 1)]

    # Convert into sum-of-products form.
    sop = SOPform([w, x, y, z], minterms, dontcares)
    # sop = sympy.simplify_logic(sop)

    # Create lambda function from boolean formula
    func = lambdify((w, x, y, z), sop)

    # Check if func really implements the truth table
    for args in product(*([[0, 1]] * 4)):
        if args not in dontcares:
            if args in minterms:
                assert func(*args)
            else:
                assert not func(*args)


def test_generate_all_n_input_formulas():
    n = 2

    inputs = [sympy.Symbol(chr(ord('a') + i)) for i in range(n)]

    truth_table_len = 2 ** n
    num_truth_tables = truth_table_len ** 2

    truth_table_input = list(product(*([[0, 1]] * n)))
    truth_tables_z = product(*([[0, 1]] * truth_table_len))

    for truth_table in truth_tables_z:

        min_terms = []
        for i, z in zip(truth_table_input, truth_table):
            if z == 1:
                min_terms.append(i)

        sop = SOPform(inputs, min_terms, dontcares=[])


def enumerate_all_multi_graphs(num_nodes, num_edges):
    """

    :param num_nodes:
    :param num_edges:
    :return:
    """

    """
    Start with g = one edge, two nodes.
    step: insert edge
        pick two existing nodes to create an edge
        if node limit not yet reached:
            pick a node split connected edges in two groups (2^n-2 possibilities) -> one node each
            connect both nodes with new edge (or pick an edge and split it??)
    """

    def backtrack(graph: nx.MultiGraph, num_nodes, num_edges):
        if num_edges == 0:
            return [graph]

        graphs = []

        new_edge_id = graph.number_of_edges() + 1
        if num_nodes > 0:
            # Create new graph by splitting an edge and inserting a new node.
            new_node_id = graph.number_of_nodes() + 1
            for a, b, key in graph.edges(keys=True):
                g2 = graph.copy()
                g2.remove_edge(a, b, key=key)
                g2.add_edge(a, new_node_id, key)
                g2.add_edge(new_node_id, b, new_edge_id)

                r = backtrack(g2, num_nodes - 1, num_edges - 1)
                graphs.extend(r)

        if num_edges > 0:
            # Add edge by connecting two existing nodes.
            node_combinations = combinations(graph.nodes, 2)
            for a, b in node_combinations:
                if a != b:
                    g2 = graph.copy()
                    g2.add_edge(a, b, new_edge_id)

                    r = backtrack(g2, num_nodes - 0, num_edges - 1)
                    graphs.extend(r)
                else:
                    pass
                    # TODO: Split node and insert edge if num_edges > 0

        return graphs

    # Start with a single edge.
    graph = nx.MultiGraph()
    graph.add_edge('a', 'b', 1)
    # Mark a, b as terminal nodes of the graph (such as vdd and output).
    graph.nodes['a']['terminal'] = 'a'
    graph.nodes['b']['terminal'] = 'b'

    graphs = backtrack(graph, num_nodes, num_edges)

    return graphs


def test_enumerate_all_multigraphs():
    graphs = enumerate_all_multi_graphs(5, 5)

    def node_match(attr1, attr2):
        return attr1.get('terminal', None) == attr2.get('terminal', None)

    # for i, ref in enumerate(graphs):
    #     iso0 = [g for g in graphs\ if nx.is_isomorphic(g, ref, node_match=node_match)]
    #     print('num isomorphic graphs to graph[%d]: %d'%(i, len(iso0)))

    print(len(graphs))



# Copyright (c) 2019-2020 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

from liberty.types import Group, Attribute
import logging
from typing import Optional, Tuple, Dict, List
import numpy as np
from ..logic.types import CombinationalOutput

logger = logging.getLogger(__name__)


def get_pin_information(cell_group: Group) -> Tuple[List[str], Dict[str, CombinationalOutput]]:
    """
    Get a list of input pins, output pins and the logic functions of output pins.
    :param cell_group:
    :return: (list of input pins, list of output pins, Dict[output pin, logic function], Dict[output pin, tri-state function])
    """
    input_pins = []
    outputs = dict()
    for pin_group in cell_group.get_groups('pin'):
        # Get pin name
        pin_name = pin_group.args[0]

        # Get direction of pin (input/output)
        direction = pin_group.get('direction', None)

        # Get boolean function of pin (for outputs).
        expr = pin_group.get_boolean_function('function')
        if expr is None:
            # Assert that for all output pins the logic function is defined.
            if direction == 'output':
                msg = f'Output pin has no function defined: {pin_name}'
                logger.warning(msg)

        # Get boolean expression for tri-state condition.
        tri_state = pin_group.get_boolean_function('three_state')
        if tri_state is not None:
            if direction != 'output':
                logger.error(f"Found non-output that with tri-state: {pin_name}")

        logger.info(f"Pin '{pin_name}' {direction} {expr}")

        # Check that pin direction is defined.
        if direction is None:
            logger.warning("Pin has undefined direction: {}/{}".format(cell_group.args[0], pin_name))

        # Remember input and output pins.
        if direction == 'input':
            input_pins.append(pin_name)
        elif direction == 'output':
            comb = CombinationalOutput(function=expr, high_impedance=tri_state)
            outputs[pin_name] = comb
        else:
            logger.warning("Pin direction type not handled: {}".format(direction))

    return input_pins, outputs


def create_table_template_if_not_exists(library: Group,
                                        table_type: str,
                                        name: str,
                                        n: int, m: int,
                                        variable1: str,
                                        variable2: Optional[str] = None,
                                        ) -> Group:
    """
    Create a lookup table template if it does not exist.

    The table name will be `'{name}_template_{n}x{m}'`.

    :param library: Library group where the template is added.
    :param table_type: Group name of the template table. One of ['lu_table_template', 'power_lut_template'].
    :param name:
    :param n: Length of index_1.
    :param m: Length of index_2
    :param variable1:
    :param variable2:
    :return:
    """

    table_types = ['lu_table_template', 'power_lut_template']
    assert table_type in table_types, f"table_type must be one of {table_type}."
    names = ['delay', 'energy', 'setup', 'hold', 'passive_energy', 'recovery', 'removal']
    assert name in names, f"name must be one of {names}."

    full_name = f"{name}_template_{n}x{m}"

    logger.debug(f"Create table template: {table_type}({full_name})")

    # Test if table already exists.
    table = library.get_groups(table_type, argument=full_name)
    if table:
        return table[0]

    else:
        # Create table.
        attributes = [
            Attribute('variable_1', variable1)
        ]
        if m > 1:
            assert variable2 is not None
            attributes.append(Attribute('variable_2', variable2))

        table = Group(table_type,
                      args=[full_name],
                      attributes=attributes
                      )

        index_1 = np.arange(n, dtype=float) + 1000

        table.set_array('index_1', index_1)
        if m > 1:
            index_2 = np.arange(m, dtype=float) + 1000
            table.set_array('index_2', index_2)

        library.groups.append(table)

        return table


def create_delay_template_table(library: Group, n: int, m: int) -> Group:
    """
    Create a `lu_table_template(delay_template_nxm)` group and add it to the library.
    :param library:
    :param n:
    :param m:
    :return:
    """
    return create_table_template_if_not_exists(
        library=library,
        table_type='lu_table_template',
        name='delay',
        n=n,
        m=m,
        variable1='total_output_net_capacitance',
        variable2='input_net_transition'
    )


def create_power_template_table(library: Group, n: int, m: int) -> Group:
    """
    Create a `power_lut_template(energy_template_nxm)` group and add it to the library.
    :param library:
    :param n:
    :param m:
    :return:
    """
    return create_table_template_if_not_exists(
        library=library,
        table_type='power_lut_template',
        name='energy',
        n=n,
        m=m,
        variable1='total_output_net_capacitance',
        variable2='input_net_transition'
    )


def create_constraint_template_table(library: Group, name: str, n: int, m: int) -> Group:
    """
    Create a `lu_table_template(delay_template_nxm)` group and add it to the library.
    :param library:
    :param name: 'setup', 'hold', 'removal' or 'recovery'
    :param n:
    :param m:
    :return:
    """
    return create_table_template_if_not_exists(
        library=library,
        table_type='lu_table_template',
        name=name,
        n=n,
        m=m,
        variable1='related_pin_transition',
        variable2='constrained_pin_transition'
    )



# Copyright (c) 2021 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

from typing import Dict, List, Set
import sympy
from sympy.logic import boolalg
from .logic.types import CombinationalOutput


class CellType:
    """
    Description of a standard-cell type.
    A `CellType` object should contain information that is necessary for the characterization of this type of cells.
    """

    def __init__(self):
        self.power_down_function: boolalg.Boolean = None
        "Boolean expression that tells when the cell is not powered."
        self.inputs: List[sympy.Symbol] = []
        "Input pins of the cell."
        self.outputs: Dict[sympy.Symbol, CombinationalOutput] = dict()
        "Dictionary with output pins mapped to their boolean expressions."

    def get_tristate_outputs(self) -> Set[sympy.Symbol]:
        """
        Return all output pins which can be high-impedance.
        :return:
        """
        return {name for name, comb in self.outputs.items() if comb.is_tristate()}

    def human_readable_description(self) -> str:
        raise NotImplementedError()


class Combinational(CellType):
    """
    Purely combinational cell without any feed-back loops.
    """

    def __init__(self):
        super().__init__()


class Latch(CellType):

    def __init__(self):
        super().__init__()

        self.data_in = None
        self.enable = None  # Write condition / clock.
        self.clear = None  # Clear condition.
        self.preset = None  # Preset condition.

    def __str__(self):
        return self.human_readable_description()

    def human_readable_description(self) -> str:
        return f"""Latch {{
    write data: {self.data_in}
    write enable: {self.enable}
    clear: {self.clear}
    preset: {self.preset}
}}"""


class SingleEdgeDFF(CellType):
    """
    Single-edge triggered delay flip-flop.
    """

    def __init__(self):
        super().__init__()

        self.internal_state: sympy.Symbol = None
        """
        Variable for the current internal state.
        """

        self.clocked_on: boolalg.Boolean = sympy.false
        "Clocked when the value of the boolean expression rises to true."
        self.next_state: boolalg.Boolean = sympy.false
        "Next state that follows a clock edge."

        self.outputs: Dict[sympy.Symbol, CombinationalOutput] = dict()
        """
        Boolean functions for all outputs. The output functions
        are functions of the primary inputs and of the `internal_state` variable.
        """

        self.scan_enable = None
        "Name of the scan-enable input."
        self.scan_in = None

        self.async_preset: boolalg.Boolean = sympy.false
        "Preset condition."

        self.async_clear: boolalg.Boolean = sympy.false
        "Clear condition."

    def __str__(self):
        return self.human_readable_description()

    def clock_signal(self) -> sympy.Symbol:
        """
        Return the clock signal if there is exactly one clock signal.
        :return:
        """
        atoms = list(self.clocked_on.atoms(sympy.Symbol))
        if len(atoms) == 1:
            return atoms[0]
        else:
            return None

    def clock_edge_polarity(self) -> bool:
        """
        Get the polarity of the clock edge if there is exactly one clock signal.
        If there are multiple or no clock signal, return `None`.
        """
        clock = self.clock_signal()
        if clock is None:
            return None
        return self.clocked_on.subs({clock: True}) == True

    def human_readable_description(self) -> str:

        return f"""SingleEdgeDFF {{
    internal_state: {self.internal_state}
    next_state: {self.next_state}
    clocked_on: {self.clocked_on}
    active clock edge: {"rising" if self.clock_edge_polarity() else "falling"}
    outputs: {self.outputs}

    asynchronous preset: {self.async_preset} 
    asynchronous clear: {self.async_clear} 

    scan enable: {self.scan_enable}
    scan input: {self.scan_in}
}}"""



# Copyright (c) 2019 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

from pyparsing import *
import re

"""
Simple SPICE parser used to read transistor netlists.
"""


class Resistor:
    def __init__(self, name, n1, n2, value=0, params={}):
        self.name = name
        self.n1 = n1
        self.n2 = n2
        self.value = value
        self.params = params

    def __repr__(self):
        return "Resistor(%s, %s, %s, %f, params=%s)" % (self.name, self.n1, self.n2, self.value, self.params)


class Capacitor:
    def __init__(self, name, n1, n2, value=0):
        self.name = name
        self.n1 = n1
        self.n2 = n2
        self.value = value

    def __repr__(self):
        return "Capacitor(%s, %s, %s, %f)" % (self.name, self.n1, self.n2, self.value)


class MOSFET:
    def __init__(self, name, nd, ng, ns, nb, model_name, params={}):
        self.name = name
        self.nd = nd
        self.ng = ng
        self.ns = ns
        self.nb = nb
        self.model_name = model_name
        self.params = params

    def __repr__(self):
        return "MOSFET(%s, %s, %s, %s, %s, %s, params=%s)" % (
            self.name, self.nd, self.ng, self.ns, self.nb, self.model_name, self.params)


class Subckt:
    def __init__(self, name, ports=[], content=[]):
        self.name = name
        self.ports = ports
        self.content = content

    def __repr__(self):
        return "Subckt(%s, %s, %s)" % (self.name, self.ports, self.content)


class Include:
    def __init__(self, path):
        self.path = path

    def __repr__(self):
        return ".include '%s'" % self.path


def parse_spice(inp):
    # Strip away comments but don't change number of lines.
    inp = re.sub(r"^\*.*$", "", inp, flags=re.MULTILINE)

    # Concatenate lines
    inp = re.sub("\n\+", " ", inp, flags=re.MULTILINE)

    # Don't ignore newline.
    ParserElement.setDefaultWhitespaceChars(' \t')

    # Elementary tokens
    beginl = (Word('\n') * (0, None)).suppress()
    endl = (LineEnd() * (1, None)).suppress()

    def parse_float(t):
        scale = 1
        if len(t) > 1:
            suffix = t[1]
            units = {
                'a': 1e-18,
                'f': 1e-15,
                'p': 1e-12,
                'n': 1e-9,
                'u': 1e-6,
                'U': 1e-6,
                'm': 1e-3,
                'k': 1e3,
                'M': 1e6,
                'G': 1e9,
                'T': 1e12
            }

            scale = units.get(suffix, 1)

        return float(t[0]) * scale

    # number = -123.123e-123
    number = (Combine(
        Optional('-') + Word(nums) + Optional('.' + Word(nums)) + Optional(Word('eE') + Optional('-') + Word(nums)))
              + Optional(Word(alphas)).leaveWhitespace()
              ).setParseAction(lambda t: parse_float(t)).setName('float')

    name = Word(alphanums + '_' + '#')
    net = name.setName('net')
    path = CharsNotIn('\n')

    # .end
    end = (beginl + '.end' + Optional('s') + Optional(name)).suppress()

    # Parameter list
    param = (name + Suppress('=') + number).setParseAction(lambda t: (t[0].upper(), t[1]))
    parameter_list = (param * (1, None)).setParseAction(lambda t: dict(list(t)))

    # Components
    resistor = (
            Combine(Word('rR') + name) + net + net + Optional(number) + Optional(parameter_list)
    ).setParseAction(lambda s, l, t: Resistor(*t))

    capacitor = (
            Combine(Word('cC') + name) + net + net + Optional(number) + Optional(parameter_list)
    ).setParseAction(lambda t: Capacitor(*t))

    mosfet = (
            Combine(Word('mM') + name) + net + net + net + net + Optional(name) + Optional(parameter_list)
    ).setParseAction(lambda t: MOSFET(*t))

    # Component
    component = (resistor | capacitor | mosfet) + endl

    # Include
    include = (Suppress('.include') + path + endl).setParseAction(lambda t: Include(*t))

    # subckt start
    subcircuit_def = (
            beginl +
            Suppress('.subckt') +
            name +
            (net * (0, None)).setParseAction(lambda t: [t]) +
            endl
    )

    # subckt body
    subcircuit = Forward()
    subcircuit << subcircuit_def + ((component | subcircuit | include) * (0, None)).setParseAction(
        lambda t: [t]) + end + endl
    subcircuit.setParseAction(lambda t: Subckt(*t))

    netlist = ((subcircuit | include) * (0, None))

    parsed = netlist.parseString(inp, parseAll=True)

    return parsed


def test_spice_parser():
    inp = """
*
*	
.subckt testCircuit in1 in2 out1
*
R1 gnd vdd 4.7 T=270K
R2 gnd vdd 6.3e3
C1 gnd vdd 0.47
* Comment
M1 gnd in1 vdd gnd nmos T=300K
.end
.include blabla
.subckt empty
.end

.subckt inv1 in out vcc gnd
MN1 out in gnd gnd NMOS L=0.35U W=2.0U
MP2 out in vcc vcc PMOS L=0.35U W=4.0U
.end

.subckt container
R1 gnd vdd 4.7
.subckt child a b c
R1 gnd vdd 4.7
.end
.end

.subckt testCircuit2 in1 in2 out1
R1 gnd vdd 4.7
R2 gnd vdd 6.3

C1 gnd vdd 0.47
C2 gnd vdd 0.47555
.end"""

    parsed = parse_spice(inp)

    print(parsed)



# Copyright (c) 2019-2020 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
Characterization functions for combinatorial cells.
"""

from typing import List, Dict, Callable, Optional

from itertools import product
import os

from .util import *
from .piece_wise_linear import *
from .ngspice_subprocess import simulate_cell
from ..lccommon.net_util import get_subcircuit_ports

import logging

logger = logging.getLogger(__name__)


def characterize_comb_cell(
        input_pins: List[str],
        output_pin: str,
        related_pin: str,
        output_functions: Dict[str, Callable],

        total_output_net_capacitance: np.ndarray,
        input_net_transition: np.ndarray,

        cell_conf: CellConfig,

        constant_inputs: Dict[str, bool] = dict(),
        initial_node_voltages: Dict[str, float] = None,

) -> Dict[str, np.ndarray]:
    """
    Calculate the NDLM timing table of a cell for a given timing arc.
    :param constant_inputs: A truth-assignment for input pins that are static. Used for tri-state enable pins.
    :param cell_conf: Parameters and configuration for the characterization.
    :param input_net_transition: Transition times of input signals in seconds.
    :param total_output_net_capacitance: Load capacitance in Farads.
    :param input_pins: List of input pins.
    :param output_pin: The output pin of the timing arc.
    :param related_pin: The input pin of the timing arc.
    :param output_functions: A dict mapping output pin names to corresponding boolean functions.
    :param initial_conditions: Initial node voltages as a dict.

    :return: Returns the NDLM timing tables wrapped in a dict:
    {'cell_rise': 2d-np.ndarray, 'cell_fall': 2d-np.ndarray, ... }
    """

    assert isinstance(cell_conf, CellConfig)
    assert isinstance(cell_conf.global_conf, CharacterizationConfig)
    cfg = cell_conf.global_conf

    inputs_inverted = cell_conf.complementary_pins.values()
    assert related_pin not in inputs_inverted, f"Active pin '{related_pin}' must not be an inverted pin of a differential pair."
    input_pins_non_inverted = [p for p in input_pins if p not in inputs_inverted]
    related_pin_inverted = cell_conf.complementary_pins.get(related_pin)

    # Find ports of the SPICE netlist.
    ports = get_subcircuit_ports(cell_conf.spice_netlist_file, cell_conf.cell_name)
    logger.debug("Subcircuit ports: {}".format(", ".join(ports)))

    vdd = cfg.supply_voltage

    # TODO
    # Maximum simulation time.
    time_max = cfg.time_step * 1e5

    # Find function to summarize different timing arcs.
    # TODO: Make this directly parametrizable by caller.
    reduction_function = {
        CalcMode.WORST: max,
        CalcMode.BEST: min,
        CalcMode.TYPICAL: np.mean
    }[cfg.timing_corner]
    
    # Initial node voltages.
    if initial_node_voltages is None:
        initial_node_voltages = dict()
    else:
        initial_node_voltages = initial_node_voltages.copy()

    # Create a list of include files.
    setup_statements = cfg.setup_statements + [f".include {cell_conf.spice_netlist_file}"]

    # Get all input nets that are not toggled during a simulation run.
    static_input_nets = [i for i in input_pins_non_inverted if i != related_pin]
    # Get a list of all input combinations that will be used for measuring conditional timing arcs.
    num_inputs = len(static_input_nets)
    static_inputs = list(product(*([[0, 1]] * num_inputs)))

    # Get boolean function of output pin.
    assert output_pin in output_functions, \
        "Boolean function not defined for output pin '{}'".format(output_pin)
    output_function = output_functions[output_pin]

    def f(input_transition_time: float, output_cap: float):
        """
        Evaluate cell timing at a single input-transition-time/output-capacitance point.
        :param input_transition_time: Transition time (slew) of the active input signal.
        :param output_cap: Load capacitance at the output.
        :return:
        """

        # Empty results.
        rise_transition_durations = []
        fall_transition_durations = []
        rise_delays = []
        fall_delays = []

        rise_powers = []
        fall_powers = []

        for static_input in static_inputs:

            # Check if the output is controllable with this static input.
            bool_inputs = {net: value > 0 for net, value in zip(static_input_nets, static_input)}

            bool_inputs[related_pin] = False
            output_when_false = output_function(**bool_inputs)
            bool_inputs[related_pin] = True
            output_when_true = output_function(**bool_inputs)

            if output_when_false == output_when_true:
                # The output will not change when this input is changed.
                # Simulation of this conditional arc can be skipped.
                logger.debug("Simulation skipped for conditional arc (output does not toggle): {}".format(bool_inputs))
                continue

            # Get voltages at static inputs.
            input_voltages = {net: cfg.supply_voltage * value for net, value in zip(static_input_nets, static_input)}
            # Add supply voltage.
            input_voltages[cell_conf.supply_net] = cfg.supply_voltage
            # Add input voltages for inverted inputs of differential pairs.
            for p in static_input_nets:
                inv = cell_conf.complementary_pins.get(p)
                if inv is not None:
                    assert inv not in input_voltages
                    # Add the inverted input voltage.
                    input_voltages[inv] = cfg.supply_voltage - input_voltages[p]

            # Add constant voltages (tri-state enable pin).
            for pin, value in constant_inputs.items():
                input_voltages[pin] = cfg.supply_voltage if value else 0.0

            logger.debug("Static input voltages: {}".format(input_voltages))

            for input_rising in [True, False]:

                # Simulation script file path.
                file_name = f"lctime_combinational_" \
                            f"slew={input_transition_time}_" \
                            f"load={output_cap}" \
                            f"{''.join((f'{net}={v}' for net, v in input_voltages.items() if isinstance(v, float)))}_" \
                            f"{'rising' if input_rising else 'falling'}"
                sim_file = os.path.join(cfg.workingdir, f"{file_name}.sp")

                # Output file for simulation results.
                sim_output_file = os.path.join(cfg.workingdir, f"{file_name}_output.txt")
                # File for debug plot of the waveforms.
                sim_plot_file = os.path.join(cfg.workingdir, f"{file_name}_plot.svg")

                bool_inputs[related_pin] = input_rising
                expected_output = output_function(**bool_inputs)
                initial_output_voltage = 0 if expected_output else vdd

                # # Get voltages at static inputs.
                # input_voltages = {net: vdd * value for net, value in zip(static_input_nets, static_input)}
                # # Add supply voltage.
                # input_voltages[supply_net] = supply_voltage
                #
                # # Add input voltages for inverted inputs of differential pairs.
                # for p in static_input_nets:
                #     inv = complementary_pins.get(p)
                #     if inv is not None:
                #         assert inv not in input_voltages
                #         # Add the inverted input voltage.
                #         input_voltages[inv] = supply_voltage - input_voltages[p]
                #
                # logger.debug("Voltages at static inputs: {}".format(input_voltages))

                # Get stimulus signal for related pin.
                input_wave = StepWave(start_time=0, polarity=input_rising,
                                      rise_threshold=0,
                                      fall_threshold=1,
                                      transition_time=input_transition_time)
                input_wave.y = input_wave.y * vdd
                input_voltages[related_pin] = input_wave
                # Get stimulus signal for the inverted pin (if any).
                if related_pin_inverted:
                    input_wave_inverted = vdd - input_wave
                    input_voltages[related_pin_inverted] = input_wave_inverted

                    # Get the breakpoint condition.
                if expected_output:
                    breakpoint_statement = f"stop when v({output_pin}) > {vdd * 0.99}"
                else:
                    breakpoint_statement = f"stop when v({output_pin}) < {vdd * 0.01}"
                breakpoint_statements = [breakpoint_statement]

                initial_node_voltages[output_pin] = initial_output_voltage
                
                simulation_title = f"Timing simulation for pin '{related_pin}', input_rising={input_rising}."

                time, voltages, currents = simulate_cell(
                    cell_name=cell_conf.cell_name,
                    cell_ports=ports,
                    input_voltages=input_voltages,
                    initial_voltages=initial_node_voltages,
                    breakpoint_statements=breakpoint_statements,
                    output_voltages=[related_pin, output_pin],
                    output_currents=[cell_conf.supply_net],
                    simulation_file=sim_file,
                    simulation_output_file=sim_output_file,
                    max_simulation_time=time_max,
                    simulation_title=simulation_title,
                    temperature=cfg.temperature,
                    output_load_capacitances={output_pin: output_cap},
                    time_step=cfg.time_step,
                    setup_statements=setup_statements,
                    ground_net=cell_conf.ground_net,
                    debug=cfg.debug,
                )

                # Retrieve data.
                supply_current = currents[cell_conf.supply_net]
                input_voltage = voltages[related_pin]
                output_voltage = voltages[output_pin]

                if cfg.debug_plots:
                    logger.debug("Create plot of waveforms: {}".format(sim_plot_file))
                    import matplotlib
                    matplotlib.use('Agg')
                    import matplotlib.pyplot as plt
                    plt.close()
                    plt.title(f"")
                    plt.plot(time, input_voltage, 'x-', label='input voltage')
                    plt.plot(time, output_voltage, label='output voltage')
                    plt.plot(time, supply_current, label='supply current')
                    plt.legend()
                    plt.savefig(sim_plot_file)
                    plt.close()

                # TODO: What unit does rise_power/fall_power have in liberty files???
                # Is it really power or energy?
                switching_power = np.mean(supply_current * vdd)
                switching_energy = switching_power * (time[-1] - time[0])

                v_thresh = 0.5 * vdd

                # Get input signal before switching and after.
                input_a = input_voltage[0] > v_thresh
                input_b = input_voltage[-1] > v_thresh
                assert input_a != input_b

                # Get output signal before switching and after.
                output_a = output_voltage[0] > v_thresh
                output_b = output_voltage[-1] > v_thresh

                # There should be an edge in the output signal.
                # Because the input signals have been chosen that way.
                assert output_a != output_b, "Supplied boolean function and simulation are inconsistent."

                # Check if output signal toggles.
                assert output_a != output_b, "Output did not toggle."

                output_rising = output_b

                # Normalize input/output such that both have a rising edge.
                input_voltage = input_voltage if input_rising else vdd - input_voltage
                output_voltage = output_voltage if output_rising else vdd - output_voltage
                # Normalize to range [0, ..., 1]
                input_voltage /= vdd
                output_voltage /= vdd

                # Check if signals are already stabilized after one `period`.
                # assert abs(input_voltage[0]) < 0.01, "Input signal not yet stable at start."
                # assert abs(1 - input_voltage[-1]) < 0.01, "Input signal not yet stable at end."
                if input_rising:
                    output_threshold = cfg.trip_points.output_threshold_rise
                else:
                    output_threshold = cfg.trip_points.output_threshold_fall
                assert abs(output_voltage[0]) <= output_threshold, "Output signal not yet stable at start."
                assert abs(1 - output_voltage[-1]) <= output_threshold, "Output signal not yet stable at end."

                # Calculate the output slew time: the time the output signal takes to change from
                # `slew_lower_threshold` to `slew_upper_threshold`.
                output_transition_duration = get_slew_time(time, output_voltage, trip_points=cfg.trip_points)

                # Calculate delay from the moment the input signal crosses `input_threshold` to the moment the output
                # signal crosses `output_threshold`.
                cell_delay = get_input_to_output_delay(time, input_voltage, output_voltage,
                                                       trip_points=cfg.trip_points)

                if output_rising:
                    rise_delays.append(cell_delay)
                    rise_transition_durations.append(output_transition_duration)
                else:
                    fall_delays.append(cell_delay)
                    fall_transition_durations.append(output_transition_duration)

                if input_rising:
                    rise_powers.append(switching_energy)
                else:
                    fall_powers.append(switching_energy)

        return (reduction_function(np.array(rise_delays)),
                reduction_function(np.array(fall_delays)),
                reduction_function(np.array(rise_transition_durations)),
                reduction_function(np.array(fall_transition_durations)),
                reduction_function(np.array(rise_powers)),
                reduction_function(np.array(fall_powers)),
                )

    f_vec = np.vectorize(f, cache=True)

    xx, yy = np.meshgrid(input_net_transition, total_output_net_capacitance)

    # Evaluate timing on the grid.
    cell_rise, cell_fall, rise_transition, fall_transition, rise_power, fall_power = f_vec(xx, yy)

    # Return the tables by liberty naming scheme.
    return {
        'total_output_net_capacitance': total_output_net_capacitance,
        'input_net_transition': input_net_transition,
        'cell_rise': cell_rise,
        'cell_fall': cell_fall,
        'rise_transition': rise_transition,
        'fall_transition': fall_transition,
        'rise_power': rise_power,
        'fall_power': fall_power
    }



# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

import tempfile
import subprocess
import numpy as np
import matplotlib.pyplot as plt
import os
import asyncio
import logging
import threading, queue
import time
from typing import List


def test_simple_simulation():
    """Run a simple transient simulation in a ngspice subprocess and retreive the results from a file."""

    sim_file = tempfile.mktemp()
    sim_output_file = tempfile.mktemp()

    spice_simulation_netlist = f"""
*
.title Simple RC circuit.

R1 VDD Y 1k
C1 Y GND 1u ic=0V
Vsrc_vdd VDD GND PWL(0 0 1ms 0V 2ms 1V)

.control
*option abstol=10e-15
*option reltol=10e-11
set filetype=ascii
* Enable output of vector names in the first line.
set wr_vecnames
tran 1ms 10ms
wrdata {sim_output_file} v(VDD) v(Y)
exit
.endc

.end
"""
    print(f"Write simulation file: {sim_file}")
    open(sim_file, 'w').write(spice_simulation_netlist)

    # Run simulation.
    ret = subprocess.run(["ngspice", sim_file])
    print(f"Subprocess return value: {ret}")
    if ret.returncode != 0:
        print(f"ngspice simulation failed: {ret}")
    assert ret.returncode == 0

    print(f"Read output data: {sim_output_file}")
    data = np.loadtxt(sim_output_file, skiprows=1)  # Skip the header.

    a_time = data[:, 0]
    a = data[:, 1]

    y_time = data[:, 2]
    y = data[:, 3]

    assert all(a_time == y_time)

    plt.plot(a_time, a, 'x-')
    plt.plot(y_time, y, 'x-')
    plt.show()

    os.remove(sim_output_file)
    os.remove(sim_file)


def test_interactive_subprocess():
    import subprocess
    p = subprocess.Popen(['ngspice', '-p'],
                         stdin=subprocess.PIPE,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE,
                         universal_newlines=True)

    p.stdin.write("a\n")
    p.stdin.flush()
    print(p.stderr.readline())

    p.stdin.write("quit\n")
    p.stdin.flush()
    print(p.stdout.readlines())
    print(p.stderr.readlines())


def test_ngspice_subprocess_class():
    sim_file = tempfile.mktemp()
    sim_output_file = tempfile.mktemp()

    spice_simulation_netlist = f"""
*
.title Simple RC circuit.

R1 VDD Y 1k
C1 Y GND 1u ic=0V
Vsrc_vdd VDD GND PWL(0 0 1ms 0V 2ms 1V)

*.control
*option abstol=10e-15
*option reltol=10e-11
*set filetype=ascii
* Enable output of vector names in the first line.
*set wr_vecnames
*tran 1ms 10ms
*wrdata {sim_output_file} v(VDD) v(Y)

*.endc

.end
    """
    print(f"Write simulation file: {sim_file}")
    open(sim_file, 'w').write(spice_simulation_netlist)

    ns = NgSpiceInteractive()
    ns.start()

    # ns.source(sim_file)
    ns.load_circuit(spice_simulation_netlist)

    # ns.cmd(spice_simulation_netlist)
    # ns.cmd('set filetype=ascii')
    # ns.cmd('tran 1ms 1ms')
    nrows = ns.tran(t_step='1ms', t_stop='1ms')
    data1 = ns.get_data(num_rows=nrows, voltages=['VDD', 'Y'], currents=['vsrc_vdd'])
    print(data1)

    nrows = ns.tran(t_step='1ms', t_stop='2ms')
    data2 = ns.get_data(num_rows=nrows, voltages=['VDD', 'Y'], currents=['vsrc_vdd'])
    print(data2)
    # ns.cmd('wrdata /dev/stdout v(VDD) v(Y)')
    ns.drop_stderr()

    err = ns.readline_err()
    print(err)

    ns.cmd('quit')


class NgSpiceAsync:
    """
    Asynchronous interface to ngspice as a subprocess.
    """

    def __init__(self, logger: logging.Logger = None):
        if logger is None:
            logger = logging.getLogger(__name__)
        self.logger = logger
        self.ngspice_cmd = "ngspice"
        self.ngspice_args = ['-p']
        "Run ngspice in interactive pipe mode."

        self.proc = None
        "ngspice process handle."

    async def start(self):
        self.proc = await asyncio.create_subprocess_exec(
            self.ngspice_cmd,
            *self.ngspice_args,
            stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )

    async def cmd(self, cmd: bytes):
        self.logger.debug(f"Run cmd: {cmd}")
        print(f"Run cmd: {cmd}")
        # cmd = cmd.encode('utf-8')
        stdout, stderr = await self.proc.communicate(input=cmd)
        self.logger.debug(f"stdout: {stdout}")
        print(f"stdout: {stdout}")
        self.logger.debug(f"stderr: {stderr}")
        print(f"stderr: {stderr}")

        print(f"return code: {self.proc.returncode}")


def test_async_interactive_subprocess():
    async def run():
        ns = NgSpiceAsync()
        await ns.start()
        await ns.cmd(b"help")
        await ns.cmd(b"help")
        await ns.cmd(b"quit")

    asyncio.run(run())


def test_async_interactive_subprocess_raw():
    import asyncio

    async def run():
        cmd = 'ngspice -p'
        proc = await asyncio.create_subprocess_exec(
            'ngspice',
            '-p',
            stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )

        stdout, stderr = await proc.communicate(input=b"help\n")
        print(stdout)
        print(stderr)
        stdout, stderr = await proc.communicate(input=b"quit\n")

        print(f'[{cmd!r} exited with {proc.returncode}]')
        if stdout:
            print(f'[stdout]\n{stdout.decode()}')
        if stderr:
            print(f'[stderr]\n{stderr.decode()}')

    asyncio.run(run())



# Copyright (c) 2019-2020 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

from itertools import chain, product, count, combinations

import networkx as nx

from typing import Any, Dict, Tuple, List, Iterable

from ..lccommon.data_types import Transistor, ChannelType

import logging

logger = logging.getLogger(__name__)


def is_unique_under_isomorphism(g: nx.Graph, references: Iterable[nx.Graph]):
    def node_match(attr1, attr2):
        return attr1.get('type', None) == attr2.get('type', None)

    return not any((nx.is_isomorphic(g, ref, node_match=node_match) for ref in references))


def enum_graphs(max_path_len: int) -> List[Tuple[nx.Graph, Dict[Any, int]]]:
    assert max_path_len >= 0

    if max_path_len == 0:
        # Create graph with maximum path length = 0.
        g = nx.Graph()
        g.add_node("source", type="source")
        return [(nx.Graph(), {"source": 0})]
    else:
        prev = enum_graphs(max_path_len - 1)

        graphs = []

        for g, distances in prev:
            d_max = max(distances.values())
            d_max_nodes = [n for n, d in distances.items() if d == d_max]
            assert len(d_max_nodes) == 1
            d_max_node = d_max_nodes[0]

            non_dmax_nodes = [n for n, d in distances.items() if d < d_max]

            name_counter = count()
            for i in range(len(non_dmax_nodes) + 1):
                for comb in combinations(non_dmax_nodes, i):

                    g_new = g.copy()
                    distances_new = distances.copy()
                    new_node = "{}_{}".format(d_max + 1, next(name_counter))
                    g_new.add_node(new_node)
                    g_new.add_edge(new_node, d_max_node)
                    distances_new[new_node] = d_max + 1

                    for n in comb:
                        g_new.add_edge(new_node, n)

                    if is_unique_under_isomorphism(g_new, (g for g,_ in graphs)):
                        graphs.append((g_new, distances_new))

        return graphs


def test_enumerate_graphs():
    import matplotlib.pyplot as plt
    graphs = enum_graphs(3)

    print("Num graphs: ", len(graphs))

    # for g, dist in graphs:
    #     print(g)
    #     nx.draw_networkx(g, with_labels=True)
    #     plt.draw()
    #     plt.show()



# Copyright (c) 2019 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

from ..lccommon.data_types import *
import networkx as nx
from typing import Tuple, List, Set, Iterable, Dict
import klayout.db as db

import logging

logger = logging.getLogger(__name__)


def get_subcircuit_ports(file: str, subckt_name: str) -> List[str]:
    """ Find port names of a subcircuit.
    :param file: Path to the spice file containing the subcircuit.
    :param subckt_name: Name of the subcircuit.
    :return: List of node names.
    """

    _, sc = load_subcircuit(file, subckt_name)

    pins = [p.name() for p in sc.each_pin()]

    return pins



class LctimeSpiceReaderDelegate(db.NetlistSpiceReaderDelegate):
    """
    """

    def __init__(self):
        self.netlist = None

    def start(self, netlist: db.Netlist):
        self.netlist = netlist

    #def control_statement(self, stmt: str):
    #    """
    #    Process unsupported control statements such as `.lib`.
    #    """
    #    splitted = stmt.split(maxsplit = 2)

    #    if len(splitted) != 2:
    #        return False

    #    cmd, args = splitted

    #    if cmd.lower() == ".lib":
    #        print("Convert `.lib` into `.include`")
    #        pass
    #    else:
    #        
    #        logger.warning(f"Control statement ignored: `{stmt}`")
    #        return False

    def parse_element(self, element_specification: str, element: str) -> db.ParseElementData:
        """
        Parse an element card.
        Ignores device parameters. They are not necessary for extracting the logic function of a CMOS cell.
        Used to work around the issue that KLayout does not deal with device parameters which are passed via subcircuit parameters.
        """

        if element.upper() == "M":
            # Parse mosfet and ignore parameters.
            tokens = element_specification.split() # Split on whitespace.

            # Merge 'parameter = value' pairs into single strings.
            args = []
            i = 0
            while i < len(tokens):
                if tokens[i] == "=":
                    args[-1] = args[-1] + "=" + tokens[i+1]
                    i += 2
                else:
                    args.append(tokens[i])
                    i += 1


            # Strip away 'parameter=value' pairs.
            args = [arg for arg in args if not "=" in arg]

            if len(args) not in [4, 5]:
                raise Exception("Mosfet is required to have 3 or 4 nets and a model.")
            
            data = db.ParseElementData()

            data.net_names = args[0:-1]
            data.model_name = args[-1]

            return data

            
        else:
            return db.NetlistSpiceReaderDelegate.parse_element(self, element_specification, element)

        
def load_netlist(path: str) -> db.Netlist:
    """
    Load a SPICE netlist.
    :param path: Path to the spice file.
    :return: Return a KLayout Netlist object.
    """
    netlist = db.Netlist()
    netlist.case_sensitive = True
    spice_reader = db.NetlistSpiceReader(LctimeSpiceReaderDelegate())
    netlist.read(path, spice_reader)

    cell_names = ", ".join(sorted({c.name for c in netlist.each_circuit()}))
    logger.debug(f"Loaded cells: '{cell_names}'")

    logger.debug("flattening netlist")
    netlist.flatten()
    
    return netlist

# def get_subcircuit(netlist: db.Netlist, circuit_name: str) -> db.Circuit:
#     """ Get a sub circuit by name (case insensitive).
#     :param circuit_name: Name of the subcircuit.
#     :return: A tuple with the netlist and the circuit. Returns none if there's no subcircuit with this name.
#     """
#
#     if not circuit_name.isupper():
#         # KLayout converts cell names to uppercase.
#         # Check if that is still true:
#         assert netlist.circuit_by_name(circuit_name) is None, "KLayout did not convert cell names to upper case."
#         logger.info(f"Convert non-upper case cellname to upper case: '{circuit_name}' -> '{circuit_name.upper()}'")
#         circuit_name = circuit_name.upper()
#
#     circuit: db.Circuit = netlist.circuit_by_name(circuit_name)
#
#     return circuit

def load_subcircuit(path: str, circuit_name: str) -> Tuple[db.Netlist, db.Circuit]:
    """ Load a sub circuit from a SPICE file.
    :param path: Path to the spice file containing the subcircuit.
    :param circuit_name: Name of the subcircuit.
    :return: A tuple with the netlist and the circuit. Returns none if there's no subcircuit with this name.
    Returns (netlist, None) when no such circuit exists.
    """

    netlist = load_netlist(path)

    circuit: db.Circuit = netlist.circuit_by_name(circuit_name)

    # Have to return the netlist too. Otherwise it is deconstructed already.
    return netlist, circuit


def extract_transistors(circuit: db.Circuit, force_lowercase: bool = False) -> Tuple[List[Transistor], Set[str]]:
    """ Load a transistor level circuit from a circuit.

    :param path: The path to the netlist.
    :param force_lowercase: Convert all net names to lower case letters.

    Returns
    -------
    Returns a list of `Transistor`s and a list of the pin names including power pins.
    (List[Transistors], pin_names)
    """

    f = lambda s: s
    if force_lowercase:
        f = lambda s: s.lower()

    if circuit is None:
        all_circuits = [c.name for c in netlist.each_circuit()]
        logger.error("No such circuit: {}".format(circuit_name))
        logger.info(f"Circuits in netlist: {all_circuits}")
        raise Exception("No such circuit: {}".format(circuit_name))

    pins = [f(p.name()) for p in circuit.each_pin()]

    def get_channel_type(s: str):
        """Determine the channel type of transistor from the model name.
        """
        if s.lower().startswith('n'):
            return ChannelType.NMOS
        return ChannelType.PMOS

    mos4 = db.DeviceClassMOS4Transistor()
    id_gate_4 = mos4.terminal_id('G')
    id_source_4 = mos4.terminal_id('S')
    id_drain_4 = mos4.terminal_id('D')

    transistors_klayout = [
        Transistor(get_channel_type(d.device_class().name),
                   f(d.net_for_terminal(id_source_4).name),
                   f(d.net_for_terminal(id_gate_4).name),
                   f(d.net_for_terminal(id_drain_4).name),
                   channel_width=d.parameter('W') * 1e-6,  # Convert into micrometers.
                   name=d.name
                   )
        for d in circuit.each_device()
        if isinstance(d.device_class(), db.DeviceClassMOS3Transistor)
           or isinstance(d.device_class(), db.DeviceClassMOS4Transistor)]

    # with open(path) as f:
    #     source = f.read()
    #
    #     ast = spice_parser.parse_spice(source)
    #
    #     match = [s for s in ast if s.name == subckt_name]
    #
    #     if len(match) < 1:
    #         raise Exception("No valid subcircuit found in file with name '%s'." % subckt_name)
    #
    #     circuit = match[0]
    #
    #     # Get transistors
    #     transistors = [
    #         Transistor(get_channel_type(t.model_name), t.ns, t.ng, t.nd, channel_width=t.params['W'], name=t.name)
    #         for t in circuit.content if type(t) is spice_parser.MOSFET
    #     ]
    #
    #     for t in transistors_klayout:
    #         print(t.channel_width)
    #
    #     for t in transistors:
    #         print(t.channel_width)
    #
    #     return transistors, circuit.ports

    return transistors_klayout, set(pins)


def load_transistor_netlist(path: str, circuit_name: str, force_lowercase: bool = False) -> Tuple[List[Transistor], Set[str]]:
    """ Load a transistor level circuit from a spice netlist.

    :param path: The path to the netlist.
    :param force_lowercase: Convert all net names to lower case letters.

    :return: Returns a list of `Transistor`s and a list of the pin names including power pins.
        (List[Transistors], pin_names)
    :raise: Raises an exception if the circuit is not found.
    """

    # Read netlist. TODO: take netlist object as argument.
    netlist, circuit = load_subcircuit(path, circuit_name)
    if circuit is None:
        raise Exception(f"Circuit not found: '{circuit_name}'")
    return extract_transistors(circuit, force_lowercase)

def is_ground_net(net: str) -> bool:
    """ Test if net is something like 'gnd' or 'vss'.
    """
    ground_nets = {0, '0', 'gnd', 'vss', 'vgnd'}
    return net.lower() in ground_nets


def is_supply_net(net: str) -> bool:
    """ Test if net is something like 'vcc' or 'vdd'.
    """
    supply_nets = {'vcc', 'vdd', 'vpwr'}
    return net.lower() in supply_nets


def is_power_net(net: str) -> bool:
    return is_ground_net(net) or is_supply_net(net)


def get_io_pins(pin_names: Iterable[str]) -> Set[str]:
    """ Get all pin names that don't look like power pins.
    """
    return {p for p in pin_names if not is_ground_net(p) and not is_supply_net(p)}


def get_cell_inputs(transistors: Iterable[Transistor]) -> Set[str]:
    """Given the transistors of a cell find the nets connected only to transistor gates.
    Will not work for transmission gates.
    """

    transistors = [t for t in transistors if t is not None]

    gate_nets = set(t.gate_net for t in transistors)
    source_and_drain_nets = set(t.source_net for t in transistors) | set(t.drain_net for t in transistors)

    # Input nets are only connected to transistor gates.
    input_nets = gate_nets - source_and_drain_nets

    return input_nets


def _transistors2graph(transistors: Iterable[Transistor]) -> nx.MultiGraph:
    """ Create a graph representing the transistor network.
        Each edge corresponds to a transistor, each node to an electrical potential.
    """
    G = nx.MultiGraph()
    for t in transistors:
        G.add_edge(t.left, t.right, t)
    assert nx.is_connected(G)
    return G


def _is_output_net(net_name, power_nets: Iterable, transistor_graph: nx.MultiGraph) -> bool:
    """
    Determine if the net is a driven output net which is the case if there is a path from the net
    to a power rail.
    :param net_name: The net to be checked.
    :param power_nets: List of available power nets ["vdd", "gnd", ...].
    :param transistor_graph:
    :return: True, iff `net_name` is a OUTPUT net. False, iff it is a INOUT net.
    """

    return any((
        nx.has_path(transistor_graph, net_name, pn)
        for pn in power_nets
    ))



# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

from enum import Enum

from copy import deepcopy
from typing import Any, Tuple


class ChannelType(Enum):
    NMOS = 1,
    PMOS = 2


class Transistor:
    """
    Abstract representation of a MOS transistor.
    """

    def __init__(self, channel_type: ChannelType,
                 source_net: str, gate_net: str, drain_net: str,
                 channel_width=None,
                 name: str = 'M?',
                 allow_flip_source_drain: bool = True
                 ):
        """
        params:
        left: Either source or drain net.
        right: Either source or drain net.
        """
        self.name = name
        self.channel_type = channel_type
        self.source_net = source_net
        self.gate_net = gate_net
        self.drain_net = drain_net

        self.channel_width = channel_width

        self.allow_flip_source_drain = allow_flip_source_drain

        # TODO
        self.threshold_voltage = None

    def flipped(self):
        """ Return the same transistor but with left/right terminals flipped.
        """

        assert self.allow_flip_source_drain, "Flipping source and drain is not allowed."

        f = deepcopy(self)
        f.source_net = self.drain_net
        f.drain_net = self.source_net

        return f

    def terminals(self) -> Tuple[Any, Any, Any]:
        """ Return a tuple of all terminal names.
        :return:
        """
        return self.source_net, self.gate_net, self.drain_net

    def __key(self):
        return self.name, self.channel_type, self.source_net, self.gate_net, self.drain_net, self.channel_width, self.threshold_voltage

    def __hash__(self):
        return hash(self.__key())

    def __eq__(x, y):
        return x.__key() == y.__key()

    def __repr__(self):
        return "({}, {}, {})".format(self.source_net, self.gate_net, self.drain_net)




# Copyright (c) 2019-2020 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
Utility functions for handling signals such as finding signal edges, measuring delays.
"""

import numpy as np
from scipy import interpolate, optimize
from enum import Enum
from collections import namedtuple
from liberty.types import *
import logging
from typing import Dict, Iterable, Optional, List
import re

logger = logging.getLogger(__name__)


class CalcMode(Enum):
    WORST = 1
    TYPICAL = 2
    BEST = 3


TripPoints = namedtuple("TripPoints", [
    "input_threshold_rise",
    "input_threshold_fall",
    "output_threshold_rise",
    "output_threshold_fall",

    "slew_lower_threshold_rise",
    "slew_upper_threshold_rise",
    "slew_lower_threshold_fall",
    "slew_upper_threshold_fall"
])


class CellConfig:
    """
    Characterization settings specific for a single cell.
    """

    def __init__(self):
        self.global_conf: CharacterizationConfig = None
        """
        General characterization configuration.
        """

        self.cell_name: str = None
        """
        Name of the cell, consistent with the name in the liberty file.
        """

        self.spice_netlist_file: str = None
        """
        Path to SPICE netlist containing the subcircuit of the cell.
        """

        self.spice_ports: List[str] = None
        """
        Ordering of pins of the spice subcircuit that describes this cell.
        """

        self.complementary_pins: Dict[str, str] = dict()
        """Mapping of non-inverting pin name to its complementary pin name of differential pairs.
        Dict[non inverting pin, inverting pin]."""

        self.workingdir: Optional[str] = None
        "Directory where simulation files are put. Best on a ram-disk."

        self.ground_net: str = 'GND'
        "Name of the ground net."

        self.supply_net: str = 'VDD'
        "Name of the power supply net."


class CharacterizationConfig:
    """
    General settings for the standard-cell characterization runs and simulations.
    """

    def __init__(self):
        self.supply_voltage: float = 0.0
        "Supply voltage in volts."

        self.trip_points: TripPoints = None
        """
        Trip-point object which specifies the voltage thresholds of the logical values.
        """

        self.timing_corner: CalcMode = None
        """
        Specify whether to take the maximum, minimum or average capacitance value. (Over all static input combinations).
        
        One of TimingCorner.WORST, TimingCorner.BEST or TimingCorner.TYPICAL
        This defines how the default timing arc is calculated from all the conditional timing arcs.
        WORST: max
        BEST: min
        TYPICAL: np.mean
        """

        self.setup_statements: List[str] = list()
        """SPICE statements that are included at the beginning of the simulation. 
        This should be used for .INCLUDE and .LIB statements."""

        self.time_step: float = 1e-12
        "Time resolution of the simulation."

        self.temperature = 27
        "Temperature of the simulated circuit."

        self.workingdir: Optional[str] = None
        "Directory where simulation files are put. Best on a ram-disk."

        self.debug: bool = False
        "Enable more verbose debugging output."

        self.debug_plots: bool = False
        "Enable more verbose debugging output such as plots of the simulations."

        # self.roll_off_factor: float = 0.10
        # """
        # OBSOLETE
        # For characterization of setup/hold times. Must be a positive and non-zero value.
        #
        # Larger values lead to smaller setup/hold window but to increased delay.
        #
        # Define how much data delay increase is tolerated.
        # The minimal delay at a sequential cell is achieved when the input signal remains
        # stable in a very wide window around the clock edge. This enlarges
        # the required setup and hold time. To make a trade-off, a degradation of the
        # delay time is accepted by the `roll_off_factor`. If the minimal delay is `d`,
        # then a delay of `d*(1+roll_off_factor)` is targeted for the characterization
        # of setup and hold times.
        # """

        self.max_pushout_time: float = 10e-12
        """
        For characterization of setup/hold times. Must be a positive and non-zero value.
        Larger values lead to smaller setup/hold window but to increased delay.
        
        Define how much increase of the clock-to-output delay is tolerated.
        
        The minimal delay at a sequential cell is achieved when the input signal remains
        stable in a very wide window around the clock edge. This enlarges
        the required setup and hold time. To make a trade-off, a degradation of the 
        delay time is accepted by the `max_pushout_time`. If the minimal delay is `d`,
        then a delay of `d + max_pushout_time` is targeted for the characterization
        of setup and hold times.
        """

        self.input_current_for_capacitance_measurement = 10000e-9  # [A]
        """
        Define the current that is pushed into an input to measure the input capacitance.
        """

        self.time_unit: float = 1e-6 # [s]
        """
        Time unit used in the liberty file.
        """
        self.capacitance_unit: float = 1e-12 # [F]
        """
        Capacitance unit used in the liberty file.
        """

    def get_units_from_liberty(self, library: Group):
        """
        Read units from liberty data.
        """
        # Units
        # TODO: choose correct unit from liberty file
        # Get the time unit used in this library.
        # Unfortunately liberty is not consistent with the format for units...
        time_unit_str = library['time_unit'].value.lower()
        assert time_unit_str.endswith('s'), "Time unit string must end on 's' for seconds."
        assert isinstance(time_unit_str, str)
        cap_unit_factor, cap_unit_str = library['capacitive_load_unit']
        if isinstance(cap_unit_str,EscapedString):
            cap_unit_str=cap_unit_str.value
        #print("cap_unit_str: "+cap_unit_str)
        assert cap_unit_str.endswith('f'), "Capacitance unit string must end on 'f' for Farads."
        assert isinstance(cap_unit_str, str)
        assert isinstance(cap_unit_factor, float)
        cap_unit_str = cap_unit_str.lower()

        time_unit_factor = float(time_unit_str[:-2])
        time_unit_str = time_unit_str[-2:]
        time_unit_prefix = time_unit_str[:1]
        cap_unit_prefix = cap_unit_str[:1]

        prefixes = {
            'm': 1e-3,  # milli
            'u': 1e-6,  # micro
            'n': 1e-9,  # nano
            'p': 1e-12,  # pico
            'f': 1e-15,  # femto
            'a': 1e-18  # atto
        }

        # Compute actual units in terms of SI units.
        self.capacitance_unit = prefixes[cap_unit_prefix] * cap_unit_factor
        self.time_unit = prefixes[time_unit_prefix] * time_unit_factor


#
# # TODO: Add type hints for Python 3.6.
# class TripPoints(NamedTuple):
#     input_threshold_rise = 0.5
#     input_threshold_fall = 0.5
#     output_threshold_rise = 0.5
#     output_threshold_fall = 0.5
#
#     slew_lower_threshold_rise = 0.2
#     slew_upper_threshold_rise = 0.8
#     slew_lower_threshold_fall = 0.2
#     slew_upper_threshold_fall = 0.8


def is_rising_edge(voltage: np.ndarray, threshold: float = 0.5) -> bool:
    """ Check if the signal rises by comparing first and last value.
    :param voltage: Signal.
    :param threshold: Decision threshold for HIGH/LOW value.
    :return: True iff the signal rises.
    """
    return voltage[0] < threshold <= voltage[-1]


def is_falling_edge(voltage: np.ndarray, threshold: float = 0.5) -> bool:
    """ Check if the signal falls by comparing first and last value.
    :param voltage: Signal.
    :param threshold: Decision threshold for HIGH/LOW value.
    :return: True iff the signal falls.
    """
    return voltage[0] > threshold >= voltage[-1]


def transition_time(voltage: np.ndarray,
                    time: np.ndarray,
                    threshold: float,
                    n: int = -1,
                    assert_one_crossing: bool = False,
                    find_falling_edges: bool = True,
                    find_rising_edges: bool = True) -> Optional[float]:
    """ Find time of the n-th event when the signal crosses the threshold.
    :param find_rising_edges: Detect rising edges. Default is `True`.
    :param find_falling_edges: Detect falling edges. Default is `True`.
    :param voltage: np.ndarray holding voltage values.
    :param time: np.ndarray holding time values.
    :param threshold:
    :param n: Selects the event if there are multiple. 0: first event, -1: last event.
    :param assert_one_crossing: If set, then assert that the signal crosses the threshold exactly once.
    :return: Time when the signal crosses the threshold for the n-th time or `None` if there's no crossing.
    """

    y_shifted = voltage - threshold

    # Find zero-crossings.
    # 0: no zero crossing here
    # 1: crossing from negative to positive
    # -1: crossing from positive to negative
    transitions = np.sign(np.diff(np.sign(y_shifted)))
    if not find_falling_edges:
        transitions[transitions == -1] = 0 # Disable falling edges.
    if not find_rising_edges:
        transitions[transitions == 1] = 0 # Disable rising edges.
    index = np.arange(len(transitions))
    # Get indices of crossings.
    transition_indices = index[transitions != 0]

    if n >= len(transition_indices):
        # There's no such crossing.
        return None

    transition_idx = transition_indices[n]

    if assert_one_crossing:
        # Count number of zero-crossings. There should be exactly one.
        assert np.sum(transitions != 0) > 0, "Signal does not cross threshold."
        assert np.sum(transitions != 0) == 1, "Signal crosses threshold multiple times."

    is_rising = transitions[transition_idx] == 1

    if not is_rising:
        # Normalize to rising edge.
        y_shifted = -y_shifted

    # Interpolate: Find zero between the both samples.
    # y1 and y2 don't have the same sign. Find the zero-crossing inbetween.
    y1 = y_shifted[transition_idx]
    y2 = y_shifted[transition_idx + 1]
    assert y1 <= 0 <= y2
    t1 = time[transition_idx]
    t2 = time[transition_idx + 1]

    dydt = (y2 - y1) / (t2 - t1)
    # 0 = y1 + delta_t * dydt
    # delta_t = -y1/dydt
    delta_t = -y1 / dydt
    t_threshold_crossing = t1 + delta_t

    return t_threshold_crossing


def get_slew_time(time: np.ndarray, voltage: np.ndarray,
                  trip_points: TripPoints) -> float:
    """
    Calculate the slew time of `voltage` signal.
    :param time: Time axis.
    :param voltage:
    :param trip_points:
    :return: Slew time in same units as `time`.
    """

    if is_falling_edge(voltage, trip_points.slew_lower_threshold_fall):
        threshold1 = trip_points.slew_upper_threshold_fall
        threshold2 = trip_points.slew_lower_threshold_fall
    elif is_rising_edge(voltage, trip_points.slew_upper_threshold_rise):
        threshold1 = trip_points.slew_lower_threshold_rise
        threshold2 = trip_points.slew_upper_threshold_rise
    else:
        assert False, "Signal has neither a rising edge nor a falling edge."

    slew = transition_time(voltage, time, threshold2) - transition_time(voltage, time, threshold1)

    assert slew >= 0, "Slew can't be negative."

    return slew


def get_input_to_output_delay(time: np.ndarray, input_signal: np.ndarray, output_signal: np.ndarray,
                              trip_points: TripPoints) -> float:
    """Calculate delay from the moment the input signal crosses `trip_points.input_threshold_{rise,fall}`
    to the moment the output signal crosses `trip_points.output_threshold_{rise,fall}`.

    Rise/fall thresholds are selected automatically depending on start and end values of the signals.

    :param time: Time axis.
    :param input_signal: Voltage of input signal (probably normalized to [0,1]).
    :param output_signal: Voltage of output signal (probably normalized to [0,1]).
    :param trip_points: TripPoints object holding threshold values.
    :return: Delay in same units as `time`.
    """

    if is_falling_edge(input_signal, trip_points.input_threshold_fall):
        i_threshold = trip_points.input_threshold_fall
    elif is_rising_edge(input_signal, trip_points.input_threshold_rise):
        i_threshold = trip_points.input_threshold_rise
    else:
        assert False, "Input signal has neither a rising edge nor a falling edge."

    if is_falling_edge(output_signal, trip_points.output_threshold_fall):
        o_threshold = trip_points.output_threshold_fall
    elif is_rising_edge(output_signal, trip_points.output_threshold_rise):
        o_threshold = trip_points.output_threshold_rise
    else:
        assert False, "Output signal has neither a rising edge nor a falling edge."

    delay = transition_time(output_signal, time, o_threshold) - transition_time(input_signal, time, i_threshold)
    assert isinstance(delay, float)
    return delay


def read_trip_points_from_liberty(library_group: Group) -> TripPoints:
    """
    Read trip points from a liberty library Group object.
    :param library_group:
    :return:
    """
    assert library_group.group_name == 'library', "Expected a `library` group but got `{}`".format(
        library_group.group_name)

    trip_points = TripPoints(
        input_threshold_rise=library_group['input_threshold_pct_rise'] * 0.01,
        input_threshold_fall=library_group['input_threshold_pct_fall'] * 0.01,
        output_threshold_rise=library_group['output_threshold_pct_rise'] * 0.01,
        output_threshold_fall=library_group['output_threshold_pct_fall'] * 0.01,

        slew_lower_threshold_rise=library_group['slew_lower_threshold_pct_rise'] * 0.01,
        slew_upper_threshold_rise=library_group['slew_upper_threshold_pct_rise'] * 0.01,
        slew_lower_threshold_fall=library_group['slew_lower_threshold_pct_fall'] * 0.01,
        slew_upper_threshold_fall=library_group['slew_upper_threshold_pct_fall'] * 0.01,
    )

    return trip_points


def find_differential_inputs_by_pattern(patterns: Iterable[str], input_pins: Iterable[str]) -> Dict[str, str]:
    """
    Find pairs of differential inputs based on user-specified patterns.
    The patterns should have the form `['PinA,PinANeg', '%_P,%_N']`.
    A comma is used to separate the non-inverting pin name from the inverting pin name.
    A '%' can be used as a placeholder.
    :param patterns: Patterns that associate non-inverting and inverting pins.
    :param input_pins: All input pin names.
    :return: Returns a dict with the mapping Dict[non-inverting pin, inverting pin]
    """
    # Store mapping of non-inverting input to inverting input.
    differential_inputs = dict()  # Dict[non-inverting, inverting]
    # Match differential inputs.

    logger.info(f"Specified differential inputs: {patterns}")
    # Split at ','
    patterns = [tuple(d.split(',', 1)) for d in patterns]
    # Convert to regex.
    patterns = [(noninv.replace('%', "(.*)"), inv) for noninv, inv in patterns]
    patterns = [(re.compile(noninv), inv) for noninv, inv in patterns]

    for pin in input_pins:
        # Try to match a pattern.
        for p, inverted_name_template in patterns:
            result = p.search(pin)
            if result:
                if len(result.groups()) == 0:
                    basename = result.group(0)
                else:
                    basename = result.group(1)
                inv_name = inverted_name_template.replace("%", basename)
                if pin in differential_inputs:
                    # Sanity check.
                    msg = f"Multiple matches for non-inverting input '{pin}'."
                    logger.error(msg)
                    raise Exception(msg)
                # Store the mapping.
                differential_inputs[pin] = inv_name

    logger.info(f"Mapping of differential inputs: {differential_inputs}")

    # Sanity check.
    if len(set(differential_inputs.keys())) != len(set(differential_inputs.values())):
        logger.error(f"Mismatch in the mapping of differential inputs.")

    return differential_inputs



# Copyright (c) 2019-2020 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

from itertools import product
from enum import Enum
import inspect


class Unate(Enum):
    NON_UNATE = 0
    NEGATIVE_UNATE = -1
    POSITIVE_UNATE = 1


def is_unate_in_xi(bool_function, param_name: str) -> Unate:
    """ Test if the boolean function is unate in input `i`.
    A function `f` is positive unate iff: f(..., x_i = 1, ...) >= f(..., x_i = 0, ...)
    A function `f` is negative unate iff: f(..., x_i = 1, ...) <= f(..., x_i = 0, ...)
    `f` is non-unate or binate if it is neither positive unate nor negative unate.
    :param bool_function:
    :param param_name: Name of input variable.
    :return: Unate.NON_UNATE, Unate.NEGATIVE_UNATE or Unate.POSITIVE_UNATE
    """

    # TODO: use sympy to do this more elegantly (satisfiability solver)

    # Get number of inputs to function.

    params = list(inspect.signature(bool_function).parameters)
    assert param_name in params, "Function does not have parameter with name '{}'.".format(param_name)
    variable_params = [p for p in params if p != param_name]
    num_inputs = len(variable_params)

    is_positive_unate = True
    is_negative_unate = True

    fixed_inputs = list(product(*([[0, 1]] * num_inputs)))

    for inp in fixed_inputs:
        inp = list(inp)

        param_values = {n: v for n, v in zip(variable_params, inp)}
        param_values[param_name] = 0

        out0 = bool_function(**param_values)
        param_values[param_name] = 1
        out1 = bool_function(**param_values)

        _is_negative_unate = out1 <= out0
        _is_positive_unate = out1 >= out0

        is_positive_unate = is_positive_unate and _is_positive_unate
        is_negative_unate = is_negative_unate and _is_negative_unate

        is_non_unate = not is_positive_unate and not is_negative_unate
        if is_non_unate:
            break

    is_non_unate = is_positive_unate == is_negative_unate
    if is_non_unate:
        return Unate.NON_UNATE
    elif is_negative_unate:
        return Unate.NEGATIVE_UNATE
    elif is_positive_unate:
        return Unate.POSITIVE_UNATE
    else:
        assert False


def test_is_unate():
    AND = lambda a, b: a and b
    NAND = lambda a, b: not (a and b)
    XOR = lambda a, b: a ^ b

    assert is_unate_in_xi(AND, 'a') == Unate.POSITIVE_UNATE
    assert is_unate_in_xi(AND, 'b') == Unate.POSITIVE_UNATE
    assert is_unate_in_xi(NAND, 'a') == Unate.NEGATIVE_UNATE
    assert is_unate_in_xi(NAND, 'b') == Unate.NEGATIVE_UNATE

    assert is_unate_in_xi(XOR, 'a') == Unate.NON_UNATE



# Copyright (c) 2019-2020 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

import networkx as nx
from itertools import product
from typing import Any, Dict, List, Iterable, Tuple
from enum import Enum

import sympy
from sympy.logic import SOPform

from lccommon.data_types import ChannelType


class Signal(Enum):
    LOW = 0
    HIGH = 1
    X = 2
    Z = 3
    SHORT = 4


def extract_output_node_from_cmos_graph(cmos_graph: nx.MultiGraph) -> Any:
    """ Find the output node of the CMOS network
        by finding the only node which connects to both n-channel and p-channel Transistors.
    :param cmos_graph: nx.MultiGraph
        The CMOS network. Edges represent transistors. Each edge key has the form (input name, ChannelType).
    :return: Returns the node of the output signal.
    """

    pn_nodes = [n for n in cmos_graph if
                len(set((ch for _, _, (_, ch) in cmos_graph.edges(n, keys=True)))) > 1
                ]
    assert len(pn_nodes) == 1, "Number of nodes connecting to n-channel and p-channel transistor must be exactly 1."
    output_node = pn_nodes[0]
    return output_node


def evaluate_cmos_graph(cmos_graph: nx.MultiGraph, vdd_node, gnd_node, output_node,
                        input_names: List,
                        inputs: Iterable[List[bool]]) -> List[Signal]:
    """ Simulate a CMOS graph using a simple switch model for the transistors.
    The must be build from a pull-up and a pull-down network.

    :param cmos_graph: Transistor network. Each edge represents a transistor. Edge key must be a tuple (input signal name, ChannelType).
    :param vdd_node: The vdd node in `cmos_graph`.
    :param gnd_node: The gnd node in `cmos_graph`.
    :param output_node: The output node in `cmos_graph`.
    :param input_names: An list of input names.
    :param inputs: Iterable[List[bool]]
        A sequence of input signal assignments. The ordering of the bools must maatch `input_names`.
    :return: Returns a list of outputs for each input assignment. Each element will be Signal.LOW, Signal.HIGH, Signal.Z or Signal.SHORT.
        Signal.Z: Neither pull-up nor pull-down network is conductive.
        Signal.SHORT: Both pull-up and pull-down network are conductive.
    """

    assert nx.is_connected(cmos_graph), '`cmos_graph` must be connected.'
    g = cmos_graph.copy()
    g.remove_node(output_node)
    assert not nx.is_connected(g), '`cmos_graph` is not a pull-up pull-down graph.'

    output_node_exp = extract_output_node_from_cmos_graph(cmos_graph)
    assert output_node == output_node_exp

    def _eval_cmos_graph(cmos_graph, input_mapping):
        g_eval = cmos_graph.copy()

        for a, b, (input_signal, channel_type) in cmos_graph.edges(keys=True):
            inp = input_mapping[input_signal]

            if channel_type == ChannelType.PMOS:
                # Invert for pmos transistors.
                inp = not inp

            if not inp:
                g_eval.remove_edge(a, b)

        is_pulling_up = nx.has_path(g_eval, vdd_node, output_node)
        is_pulling_down = nx.has_path(g_eval, gnd_node, output_node)

        if is_pulling_up and is_pulling_down:
            return Signal.SHORT

        if not is_pulling_up and not is_pulling_down:
            return Signal.Z

        if is_pulling_down:
            return Signal.LOW

        if is_pulling_up:
            return Signal.HIGH

        assert False

    outputs = []
    for inp in inputs:
        input_mapping = {k: v for k, v in zip(input_names, inp)}
        out = _eval_cmos_graph(cmos_graph, input_mapping)
        outputs.append(out)

    return outputs


def test_evaluate_cmos_graph():
    # Create CMOS network of a nand gate and check if `evaluate_cmos_graph` behaves like the nand function.
    g = nx.MultiGraph()
    g.add_edge('vdd', 'output', ('a', ChannelType.PMOS))
    g.add_edge('vdd', 'output', ('b', ChannelType.PMOS))
    g.add_edge('gnd', '1', ('a', ChannelType.NMOS))
    g.add_edge('1', 'output', ('b', ChannelType.NMOS))

    input_names = ['a', 'b']
    n = len(input_names)
    inputs = list(product(*([[0, 1]] * n)))

    # reference logic function
    def f(a, b):
        return not (a and b)

    outputs = evaluate_cmos_graph(g, 'vdd', 'gnd', 'output', input_names, inputs)

    outputs_expected = [Signal.HIGH if f(*inp) else Signal.LOW for inp in inputs]

    assert outputs == outputs_expected


def minterms_from_cmos_graph(cmos_graph: nx.MultiGraph, vdd_node, gnd_node, output_node, input_names: List) \
        -> List[Tuple]:
    """ Simulate a CMOS network with all possible inputs and return a list of inputs that lead to a HIGH output.
    :param cmos_graph:
    :param vdd_node:
    :param gnd_node:
    :param output_node:
    :param input_names:
    :return:
    """
    # Grab input names
    # input_names = [input_name for _,_,(input_name,_) in cmos_graph.edges(keys=True)]

    n = len(input_names)
    inputs = list(product(*([[0, 1]] * n)))

    outputs = evaluate_cmos_graph(cmos_graph, vdd_node, gnd_node, output_node, input_names, inputs)

    return [inp for inp, o in zip(inputs, outputs) if o == Signal.HIGH]


def cmos_graph_to_formula(cmos_graph: nx.MultiGraph, vdd_node, gnd_node, output_node,
                          input_names: List) -> sympy.Symbol:
    """
    Find the boolean formula implemented by the push-pull network `cmos_graph`.
    :param cmos_graph:
    :param vdd_node:
    :param gnd_node:
    :param output_node:
    :param input_names: Ordering of input names.
    :return: sympy.Symbol
    """
    minterms = minterms_from_cmos_graph(cmos_graph, vdd_node, gnd_node, output_node, input_names)
    dontcares = []
    input_symbols = [sympy.Symbol(n) for n in input_names]
    sop = SOPform(input_symbols, minterms, dontcares)
    sop = sympy.simplify_logic(sop)

    return sop


def test_cmos_graph_to_formula():
    # Create CMOS network of a nand gate and check if `evaluate_cmos_graph` behaves like the nand function.
    g = nx.MultiGraph()
    g.add_edge('vdd', 'output', ('a', ChannelType.PMOS))
    g.add_edge('vdd', 'output', ('b', ChannelType.PMOS))
    g.add_edge('gnd', '1', ('a', ChannelType.NMOS))
    g.add_edge('1', 'output', ('b', ChannelType.NMOS))

    input_names = ['a', 'b']

    formula = cmos_graph_to_formula(g, 'vdd', 'gnd', 'output', input_names)

    a, b = sympy.symbols('a b')
    assert formula.equals(~(a & b)), "Transformation of CMOS graph into formula failed."



# Copyright (c) 2019-2020 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
Measurement of the input capacitance by driving the input pin with a constant current.
"""

import os

from itertools import product

from .util import *
from .piece_wise_linear import *
from .ngspice_subprocess import run_simulation
from ..lccommon.net_util import get_subcircuit_ports
import logging

from scipy import interpolate

logger = logging.getLogger(__name__)


def characterize_input_capacitances(
        input_pins: List[str],
        active_pin: str,
        output_pins: List[str],
        cell_conf: CellConfig
):
    """
    Estimate the input capacitance of the `active_pin`.
    The estimation is done by simulating a constant current flowing into an input and measuring the
    time it takes for the input to go from high to low or low to high. This time multiplied by the current
    yields the transported charge which together with the voltage difference tells the capacitance.
    The measurement is done for all combinations of static inputs (all other inputs that are not measured).

    :param input_pins: List of all input pin names.
    :param active_pin: Name of the pin to be measured.
    :param output_pins: List of cell output pins.
    :param config: Parameters for the characterization.

    """

    assert isinstance(cell_conf, CellConfig)
    assert isinstance(cell_conf.global_conf, CharacterizationConfig)
    cfg = cell_conf.global_conf

    inputs_inverted = cell_conf.complementary_pins.values()
    assert active_pin not in inputs_inverted, f"Active pin '{active_pin}' must not be an inverted pin of a differential pair."
    input_pins_non_inverted = [p for p in input_pins if p not in inputs_inverted]
    active_pin_inverted = cell_conf.complementary_pins.get(active_pin)

    logger.debug("characterize_input_capacitances()")
    # Find ports of the SPICE netlist.
    ports = get_subcircuit_ports(cell_conf.spice_netlist_file, cell_conf.cell_name)
    logger.debug("Subcircuit ports: {}".format(", ".join(ports)))

    logger.debug("Ground net: {}".format(cell_conf.ground_net))
    logger.debug("Supply net: {}".format(cell_conf.supply_net))

    vdd = cfg.supply_voltage
    logger.debug("Vdd: {} V".format(vdd))

    # Create a list of include files.
    setup_statements = cfg.setup_statements + [f".include {cell_conf.spice_netlist_file}"]

    # Load include files.
    for setup in setup_statements:
        logger.debug(f"Setup statement: {setup}")
    setup_statements_string = "\n".join(setup_statements)

    # Add output load capacitance. Right now this is 0F.
    output_load_statements = "\n".join((f"Cload_{p} {p} GND 0" for p in output_pins))

    # Choose a maximum time to run the simulation.
    time_max = cfg.time_step * 1e6

    # Find function to summarize different timing arcs.
    reduction_function = {
        CalcMode.WORST: max,
        CalcMode.BEST: min,
        CalcMode.TYPICAL: np.mean
    }[cfg.timing_corner]
    logger.debug("Reduction function for summarizing multiple timing arcs: {}".format(reduction_function.__name__))

    logger.debug("Measuring input capacitance.")

    # Generate all possible input combinations for the static input pins.
    static_input_nets = [i for i in input_pins_non_inverted if i != active_pin]
    num_inputs = len(static_input_nets)

    static_inputs = list(product(*([[0, 1]] * num_inputs)))
    logger.debug(f"Number of static input combinations: {len(static_inputs)}")

    input_current = cfg.input_current_for_capacitance_measurement
    logger.debug("Input current: {}".format(input_current))

    # Loop through all combinations of inputs.
    capacitances_rising = []
    capacitances_falling = []
    for static_input in static_inputs:
        for input_rising in [True, False]:

            # Get voltages at static inputs.
            input_voltages = {net: cfg.supply_voltage * value for net, value in zip(static_input_nets, static_input)}

            # Add input voltages for inverted inputs of differential pairs.
            for p in static_input_nets:
                inv = cell_conf.complementary_pins.get(p)
                if inv is not None:
                    assert inv not in input_voltages
                    # Add the inverted input voltage.
                    input_voltages[inv] = cfg.supply_voltage - input_voltages[p]

            logger.debug("Static input voltages: {}".format(input_voltages))

            # Simulation script file path.
            file_name = f"lctime_input_capacitance_" \
                        f"{''.join((f'{net}={v}' for net, v in input_voltages.items()))}_" \
                        f"{'rising' if input_rising else 'falling'}"
            sim_file = os.path.join(cfg.workingdir, f"{file_name}.sp")

            # Output file for simulation results.
            sim_output_file = os.path.join(cfg.workingdir, f"{file_name}_output.txt")
            # File for debug plot of the waveforms.
            sim_plot_file = os.path.join(cfg.workingdir, f"{file_name}_plot.svg")

            # Switch polarity of current for falling edges.
            _input_current = input_current if input_rising else -input_current

            # Get initial voltage of active pin.
            initial_voltage = 0 if input_rising else vdd
            initial_voltage_inv = vdd - initial_voltage

            # Get the breakpoint condition.
            if input_rising:
                breakpoint_statement = f"stop when v({active_pin}) > {vdd * 0.9}"
            else:
                breakpoint_statement = f"stop when v({active_pin}) < {vdd * 0.1}"

            static_supply_voltage_statements = "\n".join(
                (f"Vinput_{net} {cell_conf.ground_net} {net} {voltage}" for net, voltage in input_voltages.items()))

            # Initial node voltages.
            initial_conditions = {
                active_pin: initial_voltage,
                cell_conf.supply_net: cfg.supply_voltage
            }
            # Add static input voltages
            initial_conditions.update(input_voltages)
            # Add initial voltage of inverted input pin (if any).
            if active_pin_inverted:
                initial_conditions[active_pin_inverted] = initial_voltage_inv

            # Create SPICE statements for the input current sources that drive the active pin.
            input_current_source_statements = [
                f"Iinput {cell_conf.ground_net} {active_pin} PULSE(0 {_input_current} 1ns 10ps 0ps 100s)"
            ]
            if active_pin_inverted is not None:
                input_current_source_statements.append(
                    f"Iinput_inv {cell_conf.ground_net} {active_pin_inverted} PULSE(0 {-_input_current} 1ns 10ps 0ps 100s)"
                )
            input_current_source_statements = "\n".join(input_current_source_statements)

            # Create ngspice simulation script.
            sim_netlist = f"""* librecell {__name__}
.title Measure input capacitance of pin {active_pin}

.option TEMP={cfg.temperature}

{setup_statements_string}

Xcircuit_under_test {" ".join(ports)} {cell_conf.cell_name}

{output_load_statements}

Vsupply {cell_conf.supply_net} {cell_conf.ground_net} {cfg.supply_voltage}

* Input current sources.
{input_current_source_statements}

* Static input voltages.
{static_supply_voltage_statements}

* Initial conditions.
* Also all voltages of DC sources must be here if they are needed to compute the initial conditions.
.ic {" ".join((f"v({net})={v}" for net, v in initial_conditions.items()))}

.control

set filetype=ascii
set wr_vecnames

* Breakpoints
{breakpoint_statement}

* Transient simulation, use initial conditions.
tran {cfg.time_step} {time_max} uic
wrdata {sim_output_file} v({active_pin}) {" ".join((f"v({p})" for p in output_pins))}
exit
.endc

.end
"""

            # Dump netlist.
            logger.debug(sim_netlist)

            # Dump simulation script to the file.
            # logger.debug(f"Write simulation netlist: {sim_file}")
            open(sim_file, "w").write(sim_netlist)

            # Run simulation.
            stdout, stderr = run_simulation(sim_file)

            # Fetch simulation results.
            # logger.debug("Load simulation output.")
            sim_data = np.loadtxt(sim_output_file, skiprows=1)

            if sim_data.ndim != 2:
                logger.error("Simulation failed. No data was written to the output file.")
                if cfg.debug:
                    logger.error(f"ngspice: {stderr}")
                assert False, "Simulation failed. No data was written to the output file."

            time = sim_data[:, 0]
            input_voltage = sim_data[:, 1]

            if cfg.debug_plots:
                logger.debug("Create plot of waveforms: {}".format(sim_plot_file))
                import matplotlib
                matplotlib.use('Agg')
                import matplotlib.pyplot as plt
                plt.close()
                plt.title(f"Measure input capacitance of pin {active_pin}.")
                plt.plot(time, input_voltage, label=active_pin)
                plt.legend()
                plt.savefig(sim_plot_file)
                plt.close()

            # Calculate average derivative of voltage by finding the slope of the line
            # through the crossing point of the voltage with the two thresholds.
            #
            # TODO: How to chose the thresholds?
            if input_rising:
                thresh1 = vdd * cfg.trip_points.slew_lower_threshold_rise
                thresh2 = vdd * cfg.trip_points.slew_upper_threshold_rise
                assert thresh1 < thresh2
            else:
                thresh1 = vdd * cfg.trip_points.slew_upper_threshold_fall
                thresh2 = vdd * cfg.trip_points.slew_lower_threshold_fall
                assert thresh1 > thresh2

            # Find transition times for both thresholds.
            transition_time1 = transition_time(input_voltage, time, threshold=thresh1, n=-1)
            transition_time2 = transition_time(input_voltage, time, threshold=thresh2, n=-1)
            assert transition_time2 > transition_time1

            # Compute deltas of time and voltage between the crossing of the two thresholds.
            f_input_voltage = interpolate.interp1d(x=time, y=input_voltage)
            dt = transition_time2 - transition_time1
            dv = f_input_voltage(transition_time2) - f_input_voltage(transition_time1)
            # dv = input_voltage[-1] - input_voltage[0]
            # dt = time[-1] - time[0]

            # Compute capacitance.
            capacitance = float(_input_current) / (float(dv) / float(dt))

            logger.debug("dV: {}".format(dv))
            logger.debug("dt: {}".format(dt))
            logger.debug("I: {}".format(input_current))
            logger.debug("Input capacitance {}: {} F".format(active_pin, capacitance))

            if input_rising:
                capacitances_rising.append(capacitance)
            else:
                capacitances_falling.append(capacitance)

    logger.debug("Characterizing input capacitances: Done")

    # Find max, min or average depending on 'reduction_function'.
    logger.debug(
        "Convert capacitances of all timing arcs into the default capacitance ({})".format(reduction_function.__name__))
    final_capacitance_falling = reduction_function(capacitances_falling)
    final_capacitance_rising = reduction_function(capacitances_rising)
    final_capacitance = reduction_function([final_capacitance_falling, final_capacitance_rising])

    return {
        'rise_capacitance': final_capacitance_falling,
        'fall_capacitance': final_capacitance_rising,
        'capacitance': final_capacitance
    }



# Copyright (c) 2019-2020 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer <code@tkramer.ch>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

from liberty.parser import parse_liberty

from liberty.types import *
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import numpy as np

import argparse
import logging

logger = logging.getLogger(__name__)


def main_plot_timing():
    """
    Command-line tool to visualize NDLM timing models in liberty files.
    :return:
    """
    parser = argparse.ArgumentParser(description='Visualize NDLM timing arrays in liberty files.')
    parser.add_argument('-l', '--liberty', required=True, metavar='LIBERTY', type=str, help='Liberty file.')
    parser.add_argument('--debug', action='store_true', help='Enable debug mode.')
    parser.add_argument('--cell', required=True, metavar='CELL_NAME', type=str, help='Cell name.')
    parser.add_argument('--pin', required=True, metavar='PIN_NAME', type=str, help='Pin name.')
    parser.add_argument('--related-pin', required=True, metavar='RELATED_PIN_NAME', type=str,
                        help='Related pin name.')
    parser.add_argument('--timing-type', required=False, default=None, metavar='TIMING_TYPE', type=str,
                        help='Value of timing_type attribute.')

    parser.add_argument('--table', required=True, metavar='TABLE_NAME', type=str,
                        help='Type of table: cell_rise, cell_fall, rise_transition, fall_transition, ...')

    # Parse arguments
    args = parser.parse_args()

    DEBUG = args.debug
    log_level = logging.DEBUG if DEBUG else logging.INFO

    logging.basicConfig(format='%(module)16s %(levelname)8s: %(message)s', level=log_level)

    lib_file = args.liberty

    logger.info("Reading liberty: {}".format(lib_file))
    with open(lib_file) as f:
        data = f.read()

    library = parse_liberty(data)

    num_cells = len([g for g in library.get_groups('cell')])
    logger.info("Number of cells: {}".format(num_cells))

    cell = select_cell(library, args.cell)
    pin = select_pin(cell, args.pin)

    timing_table = select_timing_table(pin, related_pin=args.related_pin,
                                       timing_type=args.timing_type,
                                       table_name=args.table)

    plot_timing_ndlm(library, timing_table)


def plot3d(x_axis: np.ndarray, y_axis: np.ndarray, z_data: np.ndarray,
           x_label="",
           y_label="",
           z_label=""):
    """
    Show a 3D surface plot.
    :param x_axis:
    :param y_axis:
    :param z_data:
    :param x_label:
    :param y_label:
    :param z_label:
    :return:
    """
    xx, yy = np.meshgrid(x_axis, y_axis)

    fig = plt.figure()
    ax = fig.gca(projection='3d')
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    ax.set_zlabel(z_label)

    ax.plot_surface(xx, yy, z_data)
    plt.show()


def plot_timing_ndlm(library: Group, table: Group):
    """
    Plot a NDLM timing table.
    :param library:
    :param table:
    :return:
    """

    x_axis = table.get_array('index_2')
    y_axis = table.get_array('index_1')
    z_data = table.get_array('values')

    template_name = table.args[0]
    template = library.get_group('lu_table_template', template_name)

    time_unit = library['time_unit'].value

    x_label = template['variable_2']
    y_label = template['variable_1']

    z_label = "[{}]".format(time_unit)

    plot3d(x_axis, y_axis, z_data, x_label=x_label, y_label=y_label, z_label=z_label)


def test_plot_nldm():
    import os.path
    lib_file = os.path.join(os.path.dirname(__file__), '../../test_data/freepdk45/gscl45nm.lib')

    data = open(lib_file).read()

    library = parse_liberty(data)

    cell = select_cell(library, 'INVX2')
    pin = select_pin(cell, 'Y')

    timing_table = select_timing_table(pin, 'A', 'rise_transition')

    plot_timing_ndlm(library, timing_table)
